from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from codecs import open
from os import makedirs, path
from shutil import copyfile

here = path.abspath(path.dirname(__file__))
name = "jolt"
exec(open("jolt/version.py").read())


# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

try:
    with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
        pinned_reqs = f.readlines()
except FileNotFoundError:
    pinned_reqs = []


class BuildCommand(build_py):
    def run(self):
        build_py.run(self)

        # Install additional files required by selfdeploy plugin
        if not self.dry_run:
            target_dir = path.join(self.build_lib, name, "plugins", "selfdeploy")
            makedirs(target_dir, exist_ok=True)
            for fn in ["setup.py", "README.rst"]:
                copyfile(path.join(here, fn), path.join(target_dir,fn))

setup(
    name=name,
    cmdclass={"build_py": BuildCommand},
    version=__version__,
    python_requires=">=3.8",
    description="A task executor",
    long_description=long_description,
    url="https://github.com/srand/jolt",
    author="Robert Andersson",
    author_email="srand@github.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Programming Language :: Java",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3",
    ],
    keywords=[
        "bazel",
        "build",
        "cmake",
        "conan",
        "jolt",
        "make",
        "meson",
        "msbuild",
        "xcode",
    ],
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    install_requires=pinned_reqs or [
        "bz2file",
        "click>=8.1",
        "colorama",
        "fasteners",
        "grpcio>=1.62.2",
        "jinja2",
        "keyring",
        "keyrings.alt",
        "importlib_metadata",
        "lxml",
        "multi_key_dict",
        "ninja",
        "protobuf",
        "psutil",
        "pygit2",
        "requests",
        "tqdm",
    ],
    dependency_links=[],
    extras_require={
        "allure": ["allure-python-commons"],
        "conan": ["conan<2.0"],
        "dev": ["check-manifest"],
        "doc": ["sphinx-click", "sphinx-rtd-theme"],
        "test": ["coverage"],
    },
    package_data={
        "jolt": ["**/*.sh", "**/*.xslt", "**/*.template", "**/fstree-linux-x86_64"],
    },
    entry_points={
        "console_scripts": [
            "jolt=jolt.__main__:main",
        ],
    },
)



from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectTimeout, RequestException
from urllib.parse import urlparse, urlunparse
import keyring
import getpass


from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt import tools
from jolt.error import raise_error_if, JoltError


NAME = "http"
TIMEOUT = (3.5, 27)
TIMEOUT_HEAD = (27, 27)


class Http(cache.StorageProvider):
    def __init__(self, cache):
        super(Http, self).__init__()
        self._cache = cache
        self._uri = config.get(NAME, "uri", "http://cache")
        self._uri = self._uri.rstrip("/")
        raise_error_if(not self._uri, "HTTP URI not configured")
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._disabled = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            return None

        username = config.get(NAME, "keyring.username")
        if not username:
            username = utils.read_input(NAME + " username: ")
            raise_error_if(not username, "no username configured for " + NAME)
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            raise_error_if(not password, "no password in keyring for " + NAME)
            keyring.set_password(service, username, password)

        return HTTPBasicAuth(username, password)

    def _get_url(self, artifact):
        return artifact.tools.expand(
            "{uri}/{name}/{file}",
            uri=self._uri,
            name=artifact.task.name,
            file=fs.path.basename(artifact.get_archive_path()))

    @utils.retried.on_exception((RequestException, JoltError))
    def download(self, artifact, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        url = self._get_url(artifact)
        return artifact.tools.download(
            url,
            artifact.get_archive_path(),
            exceptions=False,
            timeout=TIMEOUT)

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((RequestException))
    def upload(self, artifact, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        url = self._get_url(artifact)
        archive = artifact.get_archive()
        return artifact.tools.upload(
            archive, url,
            exceptions=False,
            auth=self._get_auth(),
            timeout=TIMEOUT)

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((RequestException))
    def location(self, artifact):
        if self._disabled:
            return False

        url = self._get_url(artifact)
        try:
            response = tools.http_session.head(url, stream=True, timeout=TIMEOUT_HEAD, auth=self._get_auth())
        except ConnectTimeout:
            self._disabled = True
            log.warning("[HTTP] failed to establish server connection, disabled")
            return False

        log.debug("[HTTP] Head ({}): {}", response.status_code, url)
        url = urlunparse(urlparse(url))
        return url if response.status_code == 200 else ''


@cache.RegisterStorage
class HttpFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("[Http] Loaded")
        return Http(cache)



import click
import grpc
import queue
from threading import Lock
import time

from google.protobuf.timestamp_pb2 import Timestamp

from jolt import cache
from jolt import cli
from jolt import colors
from jolt import config
from jolt import hooks
from jolt import loader
from jolt import log
from jolt import manifest
from jolt import common_pb2 as common_pb
from jolt import scheduler
from jolt import utils
from jolt.error import LoggedJoltError, JoltError, raise_error, raise_error_if, raise_task_error, raise_task_error_if
from jolt.graph import GraphBuilder
from jolt.scheduler import ExecutorRegistry, JoltEnvironment, NetworkExecutor, NetworkExecutorFactory, WorkerStrategy
from jolt.tasks import TaskRegistry
from jolt.options import JoltOptions
from jolt.plugins import selfdeploy
from jolt.plugins.remote_execution import log_pb2 as log_pb
from jolt.plugins.remote_execution import log_pb2_grpc as log_grpc
from jolt.plugins.remote_execution import scheduler_pb2 as scheduler_pb
from jolt.plugins.remote_execution import scheduler_pb2_grpc as scheduler_grpc
from jolt.plugins.remote_execution import worker_pb2_grpc as worker_grpc


NAME = "scheduler"
TYPE = "Remote execution"


def locked(func):
    """ Decorator to lock a method. """
    def _f(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)
    return _f


class LogHandler(object):
    """
    Log handler for executors.

    The handler is installed in the log module and sends log messages to the
    scheduler. The scheduler then forwards the messages to the client.

    The handler is installed for the duration of the task execution.
    """

    def __init__(self, stream, task):
        self.stream = stream
        self.task = task
        self.level = log.EXCEPTION

    def emit(self, record):
        """ No log messages are emitted. """
        pass

    def handle(self, record):
        """
        Handle a log record.

        The log record is formatted and sent to the scheduler.
        """
        try:
            record.message = record.msg.format(*record.args)
        except Exception:
            record.message = record.msg

        timestamp = Timestamp()
        timestamp.FromNanoseconds(int(record.created * 1000000000))

        self.stream.push(
            scheduler_pb.TaskUpdate(
                request=self.task,
                status=common_pb.TaskStatus.TASK_RUNNING,
                loglines=[
                    common_pb.LogLine(
                        context=self.task.task_id[:8],
                        level=log.level_to_pb(record.levelno),
                        time=timestamp,
                        message=record.message,
                    ),
                ]
            )
        )

    def createLock(self):
        """ Return a lock. """
        return None


class TaskCancelledException(JoltError):
    """ An exception raised when a task is cancelled by the scheduler. """
    pass


class Queue(object):
    """ A simple queue that can be used to send messages to the scheduler. """

    def __init__(self):
        self.q = queue.Queue()

    def __next__(self):
        """ Get the next item from the queue. """
        data = self.q.get()
        return data

    def push(self, item):
        """ Push an item to the queue. """
        self.q.put(item)


class RemoteExecutor(NetworkExecutor):
    """
    Executor for remotely executed tasks.

    The executor schedules the task with the scheduler and waits for the
    scheduler to respond with a task id. The executor then waits for the
    scheduler to respond with task updates. Log messages are forwarded to the
    logging system where they are formatted and emitted.

    The executor is responsible for downloading persistent artifacts from the
    cache. The executor will not download persistent artifacts unless the
    task is marked as successfully completed.

    The executor is also responsible for downloading session artifacts from the
    cache. The executor will attempt download session artifacts regardless of the
    task status. No error is raised if the download fails.

    """

    def __init__(self, factory, session, task):
        self.factory = factory
        self.session = session
        self.task = task

    def cancel(self):
        """
        Cancel the build session.

        The build session will be cancelled if the task is cancelled.
        """
        self.session.cancel()

    def download_persistent_artifacts(self, task):
        """ Download persistent artifacts from the cache. """

        for extension in task.extensions:
            self.download_persistent_artifacts(extension)
        if not task.has_artifact():
            return
        if not task.cache.download_enabled():
            return
        if not task.is_downloadable():
            return
        raise_task_error_if(
            not task.download(persistent_only=True), task,
            "Failed to download artifact")

    def download_session_artifacts(self, task):
        """ Download session artifacts from the cache. """

        for extension in task.extensions:
            self.download_session_artifacts(extension)
        if not task.has_artifact():
            return
        if not task.cache.download_enabled():
            return
        if not task.is_downloadable():
            return
        if not task.download(session_only=True):
            task.warning("Failed to download session artifact")

    def download_log(self, task):
        """ Download log and transfer lines into local logging system. """
        request = log_pb.ReadLogRequest(
            id=task.instance,
        )
        for response in self.session.logs.ReadLog(request):
            for line in response.loglines:
                log.log(
                    log.pb_to_level(line.level),
                    line.message,
                    created=line.time.ToMicroseconds() / 1000000,
                    context=line.context[:7],
                    prefix=True)

    def update_logstash(self, task):
        """ Update logstash with the task status. """
        self.task.logstash = self.session.http_uri + "/logs/" + self.task.instance

    def run(self, env):
        """ Run the task. """
        try:
            with hooks.task_run([self.task] + self.task.extensions):
                try:
                    self.run_build(env)
                except (grpc.RpcError, grpc._channel._MultiThreadedRendezvous) as rpc_error:
                    raise_task_error(self.task, rpc_error.details(), type="Scheduler error")
        except Exception as e:
            if not self.task.is_unstable:
                raise e
        finally:
            self.download_session_artifacts(self.task)

    @utils.retried.on_exception(grpc.RpcError)
    def run_build(self, env):
        """ Initialize the build session and schedule the task. """

        try:
            self.session.make_build_request()

            self.task.queued(remote=True)
            for extension in self.task.extensions:
                extension.queued(remote=True)

            request = scheduler_pb.TaskRequest(
                build_id=self.session.build_id,
                task_id=self.task.identity,
            )
            response = self.session.exec.ScheduleTask(request)

            self.update_logstash(self.task)
            self.run_task(env, response)
            self.download_persistent_artifacts(self.task)

            self.task.finished_execution(remote=True)
            for extension in self.task.extensions:
                extension.finished_execution(remote=True)

        except TaskCancelledException:
            pass

        except (grpc.RpcError, grpc._channel._MultiThreadedRendezvous) as rpc_error:
            if rpc_error.code() != grpc.StatusCode.UNAVAILABLE:
                raise_task_error(self.task, rpc_error.details(), type="Scheduler error")
            log.warning("Scheduler error: {}", rpc_error.details())
            self.session.clear_build_request()
            raise rpc_error

        except Exception as e:
            if not isinstance(e, LoggedJoltError):
                log.exception()

            if self.factory.options.mute:
                try:
                    self.download_log(self.task)
                except Exception:
                    self.task.warning("Failed to download build log")

            self.task.failed_execution(remote=True)
            for extension in self.task.extensions:
                extension.failed_execution(remote=True)

            raise e

    def run_task(self, env, response):
        """ Run the task.

        Task updates are received from the scheduler and forwarded to the
        logging system. The task is marked as running when the scheduler
        responds with a task running status.

        A change in task status is used to determine when the task has
        completed. The task is marked as completed when the scheduler
        responds with a task passed, skipped, downloaded, or uploaded status.
        An exception is raised if the scheduler responds with a task error,
        failed, unstable, or cancelled status.
        """

        last_status = common_pb.TaskStatus.TASK_QUEUED

        for progress in response:
            for line in progress.loglines:
                log.log(
                    log.pb_to_level(line.level),
                    line.message,
                    created=line.time.ToMicroseconds() / 1000000,
                    context=line.context[:7],
                    prefix=True)

            if progress.status in [common_pb.TaskStatus.TASK_RUNNING] \
               and progress.status != self.task.status():
                self.task.running_execution(remote=True)
                for extension in self.task.extensions:
                    extension.running_execution(remote=True)

            if progress.status in [common_pb.TaskStatus.TASK_QUEUED]:
                if last_status in [common_pb.TaskStatus.TASK_RUNNING]:
                    self.task.interrupted_execution(remote=True)
                    for extension in self.task.extensions:
                        extension.interrupted_execution(remote=True)

            if progress.status in [
                    common_pb.TaskStatus.TASK_PASSED,
                    common_pb.TaskStatus.TASK_DOWNLOADED,
                    common_pb.TaskStatus.TASK_UPLOADED,
                    common_pb.TaskStatus.TASK_SKIPPED,
            ]:
                break

            if progress.status in [
                    common_pb.TaskStatus.TASK_CANCELLED,
                    common_pb.TaskStatus.TASK_FAILED,
                    common_pb.TaskStatus.TASK_UNSTABLE,
            ]:
                if last_status in [common_pb.TaskStatus.TASK_QUEUED]:
                    raise TaskCancelledException()

                for error in progress.errors:
                    with self.task.task.report() as report:
                        report.add_error(
                            error.type,
                            error.location,
                            error.message,
                            error.details,
                        )
                self.task.raise_for_status()
                raise raise_error("Remote execution failed")

            if progress.status in [
                    common_pb.TaskStatus.TASK_ERROR,
            ]:
                log.log(
                    log.VERBOSE,
                    f"Host: {progress.worker.hostname}",
                    created=time.time(),
                    context=self.task.identity[:7],
                    prefix=True)

                for error in progress.errors:
                    with self.task.task.report() as report:
                        report.add_error(
                            error.type,
                            error.location,
                            error.message,
                            error.details,
                        )
                self.task.raise_for_status(log_details=not self.factory.options.mute)
                raise raise_error("Remote execution failed")

            last_status = progress.status


class RemoteSession(object):
    """
    A session with the scheduler.

    The session is responsible for establishing a connection with the scheduler,
    registering the build and creating task executors.
    """

    def __init__(self, factory):
        # Associated executor factory.
        self.factory = factory

        # Address of the scheduler.
        self.address = config.geturi(NAME, "uri", "tcp://scheduler.:9090")
        raise_error_if(self.address.scheme not in ["tcp"], "Invalid scheme in scheduler URI config: {}", self.address.scheme)
        raise_error_if(not self.address.netloc, "Invalid network address in scheduler URI config: {}", self.address.netloc)

        # URI of scheduler HTTP endpoints.
        self.http_uri = config.get(NAME, "http_uri", f"http://{self.address.netloc}")

        # GRPC channel.
        self.channel = grpc.insecure_channel(
            target=self.address.netloc,
        )

        # GRPC stub for the scheduler service.
        self.exec = scheduler_grpc.SchedulerStub(self.channel)

        # GRPC stub for the logstash service.
        self.logs = log_grpc.LogStashStub(self.channel)

        # Read build priority from config.
        # Higher priority builds will be scheduled first.
        # Default is 0.
        self.priority = config.getint(NAME, "priority", 0)

        # The build associated with this session.
        self.build = None
        self.build_id = None

        # Flag to indicate if the build has been aborted.
        self.aborted = False

        # Lock to ensure only one build is registered at a time.
        self.lock = Lock()

        # The build environment: client, workspace, etc.
        self.buildenv = None

    def initialize(self, graph):
        """ Initialize the session with the scheduler. """
        self.tasks = graph.tasks
        self.pruned = graph.pruned

    @locked
    def make_build_request(self):
        """ Create a build request with the scheduler. """

        # If a build is already registered, return.
        if self.build:
            return

        registry = ExecutorRegistry.get()

        if not self.buildenv:
            # Create a list of parameters to send to the scheduler.
            parameters = []
            for key, value in registry.get_network_parameters(None).items():
                parameters.append(common_pb.Property(key=key, value=value))

            # Add parameters from the config / command line (-c params.key).
            parameters.extend(config.export_params())

            # Create the build environment.
            self.buildenv = common_pb.BuildEnvironment(
                client=selfdeploy.get_client(),
                parameters=parameters,
                task_default_parameters=scheduler.export_task_default_params(self.tasks),
                tasks=scheduler.export_tasks(self.tasks + self.pruned),
                workspace=loader.export_workspace(self.tasks),
                loglevel=log.get_level_pb(),
                config=config.export_config(),
            )

        # Create the build request.
        req = scheduler_pb.BuildRequest(
            environment=self.buildenv,
            priority=self.priority,
            logstream=not self.factory.options.mute,
        )

        # Register the build with the scheduler.
        self.build = self.exec.ScheduleBuild(req)

        # Wait for the scheduler to respond with a build id.
        build = self.build.next()

        # Check if the build was rejected.
        if build.status == common_pb.BuildStatus.BUILD_REJECTED:
            raise_error("Build rejected by scheduler")

        # Store the build id.
        self.build_id = build.build_id

        log.info(colors.blue("Build registered with scheduler, waiting for worker"))
        return self.build

    @locked
    def clear_build_request(self):
        """ Clear the build request. Called when a build fails. """

        # Close grpc server response stream
        if self.build:
            self.build.cancel()
        self.build = None
        self.build_id = None

    def cancel(self):
        """ Send a cancel request to the scheduler. """

        # If the build has already been aborted, return.
        if self.aborted:
            return

        # If no build is registered, return.
        if not self.build:
            self.aborted = True
            return

        if not self.build_id:
            self.clear_build_request()
            return

        req = scheduler_pb.CancelBuildRequest(build_id=self.build_id)
        try:
            response = self.exec.CancelBuild(req)
            if response.status != common_pb.BuildStatus.BUILD_CANCELLED:
                log.warning("Failed to cancel build: {}", response.status)
        except grpc.RpcError as rpc_error:
            log.warning("Failed to cancel build: {}", rpc_error.details())
        finally:
            self.aborted = True

    def create_executor(self, task):
        """ Create an executor for the given task. """
        return RemoteExecutor(self.factory, self, task)


@scheduler.ExecutorFactory.Register
class RemoteExecutionFactory(NetworkExecutorFactory):
    """
    Factory for remote executors.

    Registers a build session with the scheduler and creates task executors.
    """

    def __init__(self, options):
        workers = config.getint(NAME, "workers", 1000)
        super().__init__(max_workers=workers)
        self._options = options

    @property
    def options(self):
        return self._options

    def create_session(self, graph):
        """ Create a build session in the scheduler. """
        session = RemoteSession(self)
        session.initialize(graph)
        return session

    def create(self, session, task):
        """ Create an executor for the given task. """
        return session.create_executor(task)


log.verbose("[Remote] Loaded")


@cli.cli.command(hidden=True)
@click.option("-w", "--worker", required=True, help="Worker identifier.")
@click.option("-b", "--build", required=True, help="Build identifier to enlist for.")
@click.argument("request", required=True)
@click.pass_context
def executor(ctx, worker, build, request):
    address = config.geturi(NAME, "uri", "tcp://scheduler.:9090")
    raise_error_if(address.scheme not in ["tcp"], "Invalid scheme in scheduler URI config: {}", address.scheme)
    raise_error_if(not address.netloc, "Invalid network address in scheduler URI config: {}", address.netloc)

    channel = grpc.insecure_channel(address.netloc)
    log.verbose("Waiting for GRPC channel to connect")
    grpc.channel_ready_future(channel).result()
    log.verbose("GRPC channel established: {}", address.netloc)

    sched = worker_grpc.WorkerStub(channel)

    with open(request, "rb") as f:
        request = scheduler_pb.BuildRequest()
        request.ParseFromString(f.read())

    # Set log level
    loglevel = request.environment.loglevel
    log.set_level_pb(loglevel)

    # Import protobuf build description
    manifest.ManifestExtensionRegistry.import_protobuf(request.environment)

    options = JoltOptions(
        network=False,
        local=False,
        download=config.getboolean("network", "download", True),
        upload=config.getboolean("network", "upload", True),
        keep_going=False,
        default=request.environment.task_default_parameters,
        worker=True,
        debug=False,
        salt=None,
        jobs=1)

    log.set_worker()
    log.verbose("Local build as a worker")

    tasks = loader.JoltLoader.get().load()
    for cls in tasks:
        TaskRegistry.get().add_task_class(cls)

    # Create the
    acache = cache.ArtifactCache.get(options)
    executors = ExecutorRegistry.get(options)
    strategy = WorkerStrategy(executors, acache)
    hooks.TaskHookRegistry.get(options)
    registry = TaskRegistry.get(options)

    for task in options.default:
        registry.set_default_parameters(task)

    # Build the graph of tasks
    gb = GraphBuilder(registry, acache, options=options, progress=True, buildenv=request.environment)
    dag = gb.build(request.environment.tasks.keys())

    # Enlist to execute build tasks from the scheduler
    enlist_msg = scheduler_pb.TaskUpdate(
        request=scheduler_pb.TaskRequest(build_id=build),
        worker=scheduler_pb.WorkerAllocation(id=worker, hostname=utils.hostname()),
    )

    # A queue to send updates to the scheduler
    updates = Queue()
    updates.push(enlist_msg)

    try:
        log.info("Subscribing to tasks")

        # Subscribe to tasks
        for task in sched.GetTasks(updates):
            log.set_level_pb(loglevel)

            log.info("Queuing {}", task.task_id)
            graph_task = dag.get_task_by_identity(task.task_id)
            executor = None
            status = None

            try:
                session = {}

                # Create an executor for the task
                executor = strategy.create_executor(session, graph_task)

                # Run the task
                with log.handler(LogHandler(updates, task)):
                    executor.run(JoltEnvironment(cache=acache, worker=True))

            except KeyboardInterrupt as interrupt:
                log.info("Task cancelled")

                # Send an update to the scheduler
                update = scheduler_pb.TaskUpdate(
                    request=task,
                    status=common_pb.TaskStatus.TASK_CANCELLED,
                )
                updates.push(update)

                # Interrupt sent to process, not only the running task
                raise interrupt

            except Exception:
                status = common_pb.TaskStatus.TASK_FAILED

            else:
                status = graph_task.status()

            finally:
                errors = []

                # If the task status remains queued, mark it as failed
                if status in [common_pb.TaskStatus.TASK_QUEUED]:
                    status = common_pb.TaskStatus.TASK_FAILED

                # Add errors from the task to the update sent to the scheduler
                with graph_task.task.report() as report:
                    for error in report.errors:
                        errors.append(common_pb.TaskError(
                            type=str(error.type),
                            location=str(error.location),
                            message=str(error.message),
                            details=str(error.details),
                        ))

                # Send an update to the scheduler
                update = scheduler_pb.TaskUpdate(
                    request=task,
                    status=status,
                    errors=errors,
                )
                updates.push(update)

    except Exception as e:
        log.set_level(log.EXCEPTION)
        log.exception(e)
        raise e

    log.info("Exiting")



from jolt import Download, Parameter, Resource, Task
from jolt.error import raise_task_error_if
from jolt.tasks import TaskRegistry
from jolt import attributes
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import tools
from jolt import utils

from jolt.cache import ArtifactListAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider

import contextlib
import json
from os import path
from platform import system
import tarfile


class PodmanListVariable(ArtifactListAttribute):
    pass


class PodmanImportListVariable(PodmanListVariable):
    def apply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for tar in self.items():
            try:
                tag = artifact.podman.tags[0]
            except IndexError:
                tag = artifact.tools.expand("{canonical_name}:{identity}")
            task.tools.run(
                "podman import {} {}",
                fs.path.join(artifact.path, tar), tag, output_on_error=True)
            for extra_tag in artifact.podman.tags[1:]:
                task.tools.run("podman tag {} {}", tag, extra_tag, output_on_error=True)


class PodmanLoadListVariable(PodmanListVariable):
    def apply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run(
                "podman load -i {}",
                fs.path.join(artifact.path, image), output_on_error=True)


class PodmanPullListVariable(PodmanListVariable):
    def apply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run("podman pull {}", image, output_on_error=True)


class PodmanRmiListVariable(PodmanListVariable):
    def unapply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run("podman rmi -f {}", image, output_on_error=True)


class PodmanAttributeSet(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(PodmanAttributeSet, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name == "pull":
            return PodmanPullListVariable(self._artifact, "pull")
        if name == "load":
            return PodmanLoadListVariable(self._artifact, "load")
        if name == "imprt":
            return PodmanImportListVariable(self._artifact, "imprt")
        if name == "rmi":
            return PodmanRmiListVariable(self._artifact, "rmi")
        if name == "tags":
            return PodmanListVariable(self._artifact, "tags")
        assert False, "No such podman attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class PodmanAttributeProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "podman", PodmanAttributeSet(artifact))

    def parse(self, artifact, content):
        if "podman" not in content:
            return
        for key, value in content["podman"].items():
            getattr(artifact.podman, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "podman" not in content:
            content["podman"] = {}
        for key, attrib in artifact.podman.items():
            content["podman"][key] = attrib.get_value()

    def apply(self, task, artifact):
        artifact.podman.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.podman.unapply(task, artifact)


class PodmanClient(Download):
    """ Task: Downloads and publishes the Podman command line client.

    The task will be automatically made available after importing
    ``jolt.plugins.podman``.
    """

    name = "podman/cli"
    """ Name of the task """

    arch = Parameter("x86_64", help="Host architecture")
    """ Host architecture [x86_64] """

    collect = ["podman/podman"]

    host = Parameter(system().lower(), help="Host operating system")
    """ Host operating system [autodetected] """

    url = "https://download.podman.com/{host}/static/stable/{arch}/podman-{version}.tgz"
    """ URL of binaries """

    version = Parameter("20.10.13", help="Podman version")
    """ Podman version [20.10.13] """

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append("podman")


@attributes.requires("_image")
class Container(Resource):
    """
    Resource: Starts and stops a Podman container.
    """

    arguments = []
    """ Container argument list """

    cap_adds = []
    """ A list of capabilities to add to the container """

    cap_drops = []
    """ A list of capabilities to remove from the container """

    chroot = False
    """ Use as chroot - resource consumers will execute all commands in container """

    entrypoint = None
    """ Container entrypoint """

    environment = []
    """ Environment variables """

    image = None
    """
    Image tag or Jolt task.

    If a Jolt task is specified, its artifact must export a
    metadata string named ``tag`` with the name of the image tag.
    """

    labels = []
    """ A list of container metadata labels """

    privileged = False
    """
    Start container with elevated privileges.
    """

    ports = []
    """
    A list of container ports to publish.

    Example:

    .. code-block:: python

        ports = [
            "80",
            "443:443",
        ]

    Alternatively, assign ``True`` to publish all exposed ports to random ports.
    """

    stop_timeout = 10
    """ Timeout in seconds for stopping the container .

    When stopping the container, the task will wait for the container to stop
    for the specified number of seconds before forcefully killing it.

    Default: 10 seconds.
    """

    volumes = []
    """
    A list of volumes to mount.

    By default, the cache directory and ``joltdir`` are automatically
    mounted in the container. See :attr:`volumes_default`.
    """

    volumes_default = [
        "{joltdir}:{joltdir}",
        "{joltcachedir}:{joltcachedir}",
    ]
    """
    A list of default volumes to mount.

    By default, the cache directory and ``joltdir`` are automatically
    mounted in the container. Override to disable.
    """

    user = None
    """
    Username or UID.

    Defaults to the current user.
    """

    @property
    def _arguments(self):
        return " ".join(self.arguments)

    @property
    def _cap_adds(self):
        return " ".join([utils.option("--cap-add ", cap) for cap in self.cap_adds])

    @property
    def _cap_drops(self):
        return " ".join([utils.option("--cap-drop ", cap) for cap in self.cap_drops])

    @property
    def _entrypoint(self):
        return utils.option("--entrypoint ", self.entrypoint)

    @property
    def _environment(self):
        return " ".join([utils.option("-e ", self.tools.expand(env)) for env in self.environment])

    @property
    def _image(self):
        registry = TaskRegistry.get()
        tool = tools.Tools(self)
        if registry.get_task_class(tool.expand(self.image)):
            return [self.image]
        return []

    def _info(self, fmt, *args, **kwargs):
        """
        Log information about the task.
        """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.info(fmt, *args, **kwargs)

    @property
    def _labels(self):
        return " ".join([utils.option("-l ", self.tools.expand(label)) for label in self.labels])

    @property
    def _privileged(self):
        return "--privileged" if self.privileged else ""

    @property
    def _ports(self):
        if self.ports is True:
            return "-P"
        return " ".join([utils.option("-p ", self.tools.expand(port)) for port in self.ports])

    @property
    def _user(self):
        if self.user:
            return f"--user {self.user}"
        try:
            from os import getuid
            return "--user " + str(getuid())
        except ImportError:
            return ""

    @property
    def _volumes(self):
        return " ".join([utils.option("-v ", self.tools.expand(vol))
                         for vol in self.volumes_default + self.volumes])

    def acquire(self, artifact, deps, tools, owner):
        self.joltcachedir = config.get_cachedir()
        try:
            image = deps[self.image]
            image = str(image.strings.tag)
        except Exception:
            image = tools.expand(self.image)

        self._info(f"Creating container from image '{image}'")
        self.container = tools.run(
            "podman run -i -d {_cap_adds} {_cap_drops} {_entrypoint} {_labels} {_ports} {_privileged} {_user} {_environment} {_volumes} {image} {_arguments}",
            image=image, output_on_error=True)

        self._info("Created container '{container}'")
        info = tools.run("podman inspect {container}", output_on_error=True)
        artifact.container = self.container
        artifact.info = json.loads(info)[0]

        if self.chroot:
            self._context_stack = contextlib.ExitStack()
            self._context_stack.enter_context(
                owner.tools.runprefix(f"podman exec -i {artifact.container}"))

    def release(self, artifact, deps, tools, owner):
        if self.chroot:
            self._context_stack.close()

        try:
            self._info("Stopping container '{container}'")
            tools.run("podman stop -t {stop_timeout} {container}", output_on_error=True)
        finally:
            self._info("Deleting container '{container}'")
            tools.run("podman rm -f {container}", output_on_error=True)


class PodmanLogin(Resource):
    """
    Resource: Logs in and out of a Podman Registry.

    If the user and password parameters are unset, credentials
    are fetched from the environment variables:

        - PODMAN_USER
        - PODMAN_PASSWD

    The resource will be automatically made available after importing
    ``jolt.plugins.podman``.
    """
    name = "podman/login"
    """ Name of the resource """

    requires = ["podman/cli"]

    user = Parameter("", help="Podman Registry username")
    """
    Podman Registry username.

    If not set, the environment variable ``PODMAN_USER`` is read instead.
    """

    passwd = Parameter("", help="Podman Registry password")
    """
    Podman Registry password.

    If not set, the environment variable ``PODMAN_PASSWD`` is read instead.
    """

    server = Parameter("", help="Podman Registry server")
    """
    Podman Registry server.

    If no server is specified, the default is defined by the daemon.
    """

    def _user(self, tools):
        return str(self.user) or tools.getenv("PODMAN_USER")

    def _password(self, tools):
        return str(self.passwd) or tools.getenv("PODMAN_PASSWD")

    def acquire(self, artifact, deps, tools, owner):
        raise_task_error_if(not self._user(tools), self, "Username has not been configured")
        raise_task_error_if(not self._password(tools), self, "Password has not been configured")

        with tools.cwd(tools.builddir()):
            tools.write_file("podman-credential", self._password(tools))
            tools.run("cat podman-credential | podman login -u {user} --password-stdin {server}", user=self._user(tools))

    def release(self, artifact, deps, tools, owner):
        tools.run("podman logout {server}")


TaskRegistry.get().add_task_class(PodmanClient)
TaskRegistry.get().add_task_class(PodmanLogin)


class _Tarfile(tarfile.TarFile):

    def without(self, targetpath):
        dirname, filename = fs.path.split(targetpath)
        if not filename.startswith(".wh."):
            return None
        fs.unlink(fs.path.join(dirname, filename[4:]), ignore_errors=True, tree=True)
        return True

    def makedev(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makedev(tarinfo, targetpath)

    def makedir(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makedir(tarinfo, targetpath)

    def makefifo(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makefifo(tarinfo, targetpath)

    def makefile(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        if fs.path.lexists(targetpath):
            fs.unlink(targetpath)
        super().makefile(tarinfo, targetpath)

    def makelink(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makelink(tarinfo, targetpath)


class ContainerImage(Task):
    """
    Abstract Task: Builds and publishes a Podman image.

    Builds the selected ``Dockerfile`` and optionally tags and pushes the
    image to a registry. The image can also be saved to file and published
    in the task artifact. Compression formats supported are bzip2, gzip and
    lzma.

    By default, base images referenced in the ``Dockerfile`` will be pulled
    during the build. Note that Jolt has no way of knowing beforehand if
    images have been updated in the registry. Use time-based influence to
    trigger rebuilds if it's important that base images are kept up-to-date.

    No automatic influence for ``Dockerfile`` or context is collected. Make
    sure to use an appropriate influence decorator.

    Optionally add requirements to:

      - ``podman/cli`` to provision the Podman client, if none is available on the host.
      - ``podman/login`` to automatically login to the Podman registry.

    This class must be subclassed.

    Example:

    .. code-block:: podman

        # Dockerfile

        FROM busybox:latest
        CMD ["busybox"]

    .. code-block:: python

        # build.jolt

        from jolt.plugins.podman import ContainerImage

        class Busybox(ContainerImage):
            \"\"\" Publishes Busybox image as gzip-compressed tarball \"\"\"
            compression = "gz"
            requires = ["podman/cli"]
            tags = ["busybox:{identity}"]

    """
    abstract = True

    autoload = True
    """
    Automatically load image file into local registry when the artifact is
    consumed by another task.

    If the built image is saved to a file (i.e. ``imagefile`` is set), the image
    file is automatically loaded into the local Podman registry when the task
    artifact is consumed by another task. The image is also automatically
    removed from the registry upon completion of the consumer task.

    Default: ``True``.
    """

    buildargs = []
    """
    List of build arguments and their values ("ARG=VALUE").

    The arguments are passed to Podman using ``--build-arg``.
    """

    cleanup = True
    """ Remove local image upon completion [True] """

    context = "."
    """ Path to build context, relative to joltdir (directory). """

    dockerfile = "Dockerfile"
    """ Path to the Dockerfile to build, or the full source code of such a file. """

    output = "oci-archive"
    """ none, oci-archive, oci-directory, docker-archive, directory, tar, tar.gz, tgz, tar.bz2, tar.xz"""

    imagefile = "{canonical_name}.tar"
    """
    Name of the image tarball published by the task.

    If set to None, no image file will be saved and published.

    Defaults to the task's canonical name.
    """

    labels = []
    """ A list of image metadata labels """

    pull = True
    """
    Always pull images when building.

    Passes --pull to the Podman client.
    """

    push = False
    """
    Optionally push image to registry [False]

    To be able to push images, the current user must login to the Podman Registry.
    The ``podman/login`` Jolt resource can be used for that purpose.
    """

    squash = False
    """ Squash image layers """

    tags = ["{canonical_name}:{identity}"]
    """ Optional list of image tags. Defaults to task's canonical name. """

    target = None
    """ Target platform, e.g. linux/arm/v7. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _buildargs(self):
        return " ".join([utils.option("--build-arg ", self.tools.expand(ba)) for ba in self.buildargs])

    @property
    def _labels(self):
        return " ".join([utils.option("-l ", self.tools.expand(label)) for label in self.labels])

    @property
    def _platform(self):
        platform = self.tools.expand(self.target) if self.target else None
        return utils.option("--platform ", platform)

    @property
    def _tags(self):
        return " ".join([utils.option("-t ", tag) for tag in self.tags])

    def run(self, deps, tools):
        context = tools.expand_relpath(self.context, self.joltdir)
        dockerfile = tools.expand_path(self.dockerfile)
        self._imagefile = tools.expand(self.imagefile) if self.imagefile else None
        self._autoload = self._imagefile and self.autoload
        self.tags = [self.tools.expand(tag) for tag in self.tags]
        pull = " --pull" if self.pull else ""
        squash = " --squash" if self.squash else ""

        # If dockerfile is not relative to joltdir, look for it in context
        if not path.exists(dockerfile):
            with tools.cwd(context):
                dockerfile = tools.expand_path(self.dockerfile)

        if not path.exists(dockerfile):
            with tools.cwd(tools.builddir()):
                tools.write_file("Dockerfile", self.dockerfile)
                dockerfile = tools.expand_path("Dockerfile")

        self.info("Building image from {} in {}",
                  tools.expand_relpath(dockerfile),
                  tools.expand_relpath(context))

        with tools.cwd(context):
            tools.run("podman build {_platform} . -f {} {_buildargs} {_labels} {_tags} {pull}{squash}",
                      utils.quote(dockerfile), pull=pull, squash=squash)

        try:
            if self.push:
                self.info("Pushing image")
                for tag in self.tags:
                    tools.run("podman push {}", tag)

            if self.output:
                self.info("Saving image")
                with tools.cwd(tools.builddir()):
                    if self.output in ["oci-archive", "docker-archive"]:
                        tools.run("podman image save --format={output} {} -o {}", self.tags[0], "image.tar")
                    if self.output == "oci-directory":
                        tools.run("podman image save --format=oci-dir {} -o {}", self.tags[0], "image.dir")
                    if self.output in ["archive", "directory"]:
                        ctr = tools.run("podman create {}", self.tags[0])
                        try:
                            mount_path = tools.run("podman unshare podman mount {}", ctr, output_on_error=True)
                            if self.output == "archive":
                                tools.run("podman unshare tar -C {} -cf image.tar .", mount_path, output_on_error=True)
                            else:
                                tools.mkdir("image.dir")
                                tools.run("podman unshare tar c -C {} . | tar --no-same-permissions --no-same-owner --no-overwrite-dir -x -C ./image.dir/", mount_path, output_on_error=True)
                        finally:
                            utils.call_and_catch(tools.run, "podman rm {}", ctr)
        finally:
            if self.cleanup:
                self.info("Removing image from Podman")
                for tag in self.tags:
                    utils.call_and_catch(tools.run("podman rmi -f {}", tag))

    def publish(self, artifact, tools):
        artifact.strings.tag = tools.expand(self.tags[0])

        for tag in self.tags:
            artifact.podman.tags.append(tag)

        if self.output:
            with tools.cwd(tools.builddir()):
                if self.output in ["oci-archive", "docker-archive"] and self._imagefile:
                    artifact.collect("image.tar", "{_imagefile}")
                    if self._autoload:
                        artifact.podman.load.append("{_imagefile}")
                        artifact.podman.rmi.append(artifact.strings.tag.get_value())
                if self.output in ["archive"] and self._imagefile:
                    artifact.collect("image.tar", "{_imagefile}")
                    if self._autoload:
                        artifact.podman.imprt.append("{_imagefile}")
                        artifact.podman.rmi.append(artifact.strings.tag.get_value())
                if self.output in ["directory", "oci-directory"]:
                    with tools.cwd("image.dir"):
                        artifact.collect("*", symlinks=True)
                if self.output in ["directory"]:
                    artifact.paths.rootfs = "."



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11jolt/common.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"-\n\x06\x44igest\x12\x16\n\talgorithm\x18\x01 \x01(\tR\x03\x61lg\x12\x0b\n\x03hex\x18\x02 \x01(\t\"o\n\x07LogLine\x12\x18\n\x05level\x18\x01 \x01(\x0e\x32\t.LogLevel\x12(\n\x04time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x63ontext\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t\"&\n\x08Property\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\")\n\x08Platform\x12\x1d\n\nproperties\x18\x01 \x03(\x0b\x32\t.Property\"\x83\x01\n\x04Task\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08identity\x18\x02 \x01(\t\x12\x10\n\x08instance\x18\x03 \x01(\t\x12\r\n\x05taint\x18\x04 \x01(\t\x12\x1d\n\nproperties\x18\x05 \x03(\x0b\x32\t.Property\x12\x1b\n\x08platform\x18\x06 \x01(\x0b\x32\t.Platform\"M\n\tTaskError\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x10\n\x08location\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x04 \x01(\t\"5\n\x04\x46ile\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0e\n\x06\x64igest\x18\x02 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\"\xf1\x01\n\x07Project\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\"\n\x05paths\x18\x02 \x03(\x0b\x32\x13.Project.SystemPath\x12 \n\x07recipes\x18\x03 \x03(\x0b\x32\x0f.Project.Recipe\x12$\n\tresources\x18\x04 \x03(\x0b\x32\x11.Project.Resource\x1a\x1a\n\nSystemPath\x12\x0c\n\x04path\x18\x01 \x01(\t\x1a\'\n\x08Resource\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\x1a\'\n\x06Recipe\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0f\n\x07workdir\x18\x03 \x01(\t\"\x8e\x01\n\tWorkspace\x12\x0f\n\x07rootdir\x18\x01 \x01(\t\x12\x10\n\x08\x63\x61\x63hedir\x18\x02 \x01(\t\x12\x14\n\x05\x66iles\x18\x03 \x03(\x0b\x32\x05.File\x12\x1a\n\x08projects\x18\x04 \x03(\x0b\x32\x08.Project\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x0c\n\x04tree\x18\x06 \x01(\t\x12\x10\n\x08\x62uilddir\x18\x07 \x01(\t\"i\n\x06\x43lient\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0c\n\x04tree\x18\x02 \x01(\t\x12\x10\n\x08identity\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12\x14\n\x0crequirements\x18\x05 \x03(\t\x12\x0b\n\x03nix\x18\x06 \x01(\x08\"\x99\x02\n\x10\x42uildEnvironment\x12\x17\n\x06\x63lient\x18\x01 \x01(\x0b\x32\x07.Client\x12\x1d\n\tworkspace\x18\x02 \x01(\x0b\x32\n.Workspace\x12\x1d\n\nparameters\x18\x03 \x03(\x0b\x32\t.Property\x12\x1f\n\x17task_default_parameters\x18\x04 \x03(\t\x12+\n\x05tasks\x18\x05 \x03(\x0b\x32\x1c.BuildEnvironment.TasksEntry\x12\x1b\n\x08loglevel\x18\x06 \x01(\x0e\x32\t.LogLevel\x12\x0e\n\x06\x63onfig\x18\x07 \x01(\t\x1a\x33\n\nTasksEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x14\n\x05value\x18\x02 \x01(\x0b\x32\x05.Task:\x02\x38\x01*k\n\x08LogLevel\x12\r\n\tEXCEPTION\x10\x00\x12\t\n\x05\x44\x45\x42UG\x10\x01\x12\x0b\n\x07VERBOSE\x10\x02\x12\n\n\x06STDOUT\x10\x03\x12\x08\n\x04INFO\x10\x04\x12\x0b\n\x07WARNING\x10\x05\x12\t\n\x05\x45RROR\x10\x06\x12\n\n\x06STDERR\x10\x07*_\n\x0b\x42uildStatus\x12\x12\n\x0e\x42UILD_ACCEPTED\x10\x00\x12\x12\n\x0e\x42UILD_REJECTED\x10\x01\x12\x13\n\x0f\x42UILD_COMPLETED\x10\x03\x12\x13\n\x0f\x42UILD_CANCELLED\x10\x04*\xd4\x01\n\nTaskStatus\x12\x0f\n\x0bTASK_QUEUED\x10\x00\x12\x10\n\x0cTASK_RUNNING\x10\x01\x12\x0f\n\x0bTASK_FAILED\x10\x02\x12\x0f\n\x0bTASK_PASSED\x10\x03\x12\x11\n\rTASK_UNSTABLE\x10\x04\x12\x13\n\x0fTASK_DOWNLOADED\x10\x05\x12\x11\n\rTASK_UPLOADED\x10\x06\x12\x10\n\x0cTASK_SKIPPED\x10\x07\x12\x12\n\x0eTASK_CANCELLED\x10\x08\x12\x0e\n\nTASK_ERROR\x10\t\x12\x10\n\x0cTASK_CREATED\x10\nB\x0eZ\x0cpkg/protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.common_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\014pkg/protocol'
  _globals['_BUILDENVIRONMENT_TASKSENTRY']._options = None
  _globals['_BUILDENVIRONMENT_TASKSENTRY']._serialized_options = b'8\001'
  _globals['_LOGLEVEL']._serialized_start=1345
  _globals['_LOGLEVEL']._serialized_end=1452
  _globals['_BUILDSTATUS']._serialized_start=1454
  _globals['_BUILDSTATUS']._serialized_end=1549
  _globals['_TASKSTATUS']._serialized_start=1552
  _globals['_TASKSTATUS']._serialized_end=1764
  _globals['_DIGEST']._serialized_start=54
  _globals['_DIGEST']._serialized_end=99
  _globals['_LOGLINE']._serialized_start=101
  _globals['_LOGLINE']._serialized_end=212
  _globals['_PROPERTY']._serialized_start=214
  _globals['_PROPERTY']._serialized_end=252
  _globals['_PLATFORM']._serialized_start=254
  _globals['_PLATFORM']._serialized_end=295
  _globals['_TASK']._serialized_start=298
  _globals['_TASK']._serialized_end=429
  _globals['_TASKERROR']._serialized_start=431
  _globals['_TASKERROR']._serialized_end=508
  _globals['_FILE']._serialized_start=510
  _globals['_FILE']._serialized_end=563
  _globals['_PROJECT']._serialized_start=566
  _globals['_PROJECT']._serialized_end=807
  _globals['_PROJECT_SYSTEMPATH']._serialized_start=699
  _globals['_PROJECT_SYSTEMPATH']._serialized_end=725
  _globals['_PROJECT_RESOURCE']._serialized_start=727
  _globals['_PROJECT_RESOURCE']._serialized_end=766
  _globals['_PROJECT_RECIPE']._serialized_start=768
  _globals['_PROJECT_RECIPE']._serialized_end=807
  _globals['_WORKSPACE']._serialized_start=810
  _globals['_WORKSPACE']._serialized_end=952
  _globals['_CLIENT']._serialized_start=954
  _globals['_CLIENT']._serialized_end=1059
  _globals['_BUILDENVIRONMENT']._serialized_start=1062
  _globals['_BUILDENVIRONMENT']._serialized_end=1343
  _globals['_BUILDENVIRONMENT_TASKSENTRY']._serialized_start=1292
  _globals['_BUILDENVIRONMENT_TASKSENTRY']._serialized_end=1343
# @@protoc_insertion_point(module_scope)



from jolt.cache import ArtifactListAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider


class CppInfoListVariable(ArtifactListAttribute):
    pass


class CppInfoDictVariable(ArtifactListAttribute):
    def __init__(self, artifact, name):
        super(CppInfoDictVariable, self).__init__(artifact, name)

    def __setitem__(self, key, value=None):
        item = "{0}={1}".format(key, value) if value is not None else key
        self.append(item)


class CppInfo(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(CppInfo, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name == "asflags":
            return CppInfoListVariable(self._artifact, "asflags")
        if name == "cflags":
            return CppInfoListVariable(self._artifact, "cflags")
        if name == "cxxflags":
            return CppInfoListVariable(self._artifact, "cxxflags")
        if name == "incpaths":
            return CppInfoListVariable(self._artifact, "incpaths")
        if name == "ldflags":
            return CppInfoListVariable(self._artifact, "ldflags")
        if name == "libpaths":
            return CppInfoListVariable(self._artifact, "libpaths")
        if name == "libraries":
            return CppInfoListVariable(self._artifact, "libraries")
        if name == "macros":
            return CppInfoDictVariable(self._artifact, "macros")
        if name == "sources":
            return CppInfoListVariable(self._artifact, "sources")
        assert False, "No such cxxinfo attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class CppInfoProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "cxxinfo", CppInfo(artifact))

    def parse(self, artifact, content):
        if "cxxinfo" not in content:
            return
        for key, value in content["cxxinfo"].items():
            getattr(artifact.cxxinfo, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "cxxinfo" not in content:
            content["cxxinfo"] = {}
        for key, attrib in artifact.cxxinfo.items():
            content["cxxinfo"][key] = attrib.get_value()

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass
